import sys
import os
import os.path
import ac
import acsys 
import time
import struct
import json
import datetime
import threading
from math import *
import subprocess

# import dlls 
import platform
if platform.architecture()[0] == "64bit":
    ddlDir=os.path.dirname(__file__)+'/DLLsx64'
else:
    ddlDir=os.path.dirname(__file__)+'/DLLs'
sys.path.insert(0, ddlDir)
os.environ['PATH'] = os.environ['PATH'] + ";."
import winsound
import socketserver
import socket


################################################################################
#### Constant
################################################################################

### Methods
NEW_APP = 1
SET_SIZE = 2
SET_TITLE = 3
SET_BACKGROUND_TEXTURE = 4
SET_VISIBLE = 5
ADD_LABEL = 6
SET_POSITION = 7
DRAW_BACKGROUND = 8
SET_BACKGROUND_COLOR = 9
SET_BACKGROUND_OPACITY = 10
ADD_BUTTON = 11
SET_FONT_ALIGNMENT = 12
SET_FONT_COLOR = 13
SET_FONT_SIZE = 14
SET_TEXT = 15
GET_TRACK_CONFIGURATION = 30
CUSTOM1 = 100
CUSTOM2 = 101
CUSTOM3 = 102
CUSTOM4 = 103
CUSTOM5 = 104
CUSTOM6 = 105
CUSTOM7 = 106
CUSTOM8 = 107
CUSTOM9 = 108
CUSTOM10 = 109

### Events
ON_APP_ACTIVATED = 1
ON_APP_DISMISSED = 2
ON_BUTTON_CLICK = 3
ON_APP_INITIALIZED = 4
ON_SHUTDOWN = 5
ON_GET_TRACK_CONFIGURATION = 6
ON_FALSE_START = 7

ON_30_SEC = 8
ON_15_SEC = 9
ON_10_SEC = 10
ON_5_SEC = 11

ON_POSITION = 12

###############################################################################
### Global var
###############################################################################
server = None
acObj = {}  # Dictionnary : pair key / Value
iButtonClick = 0
buttonList = []
appWindow = None
lightTexture = -1
isDisplayChanged = True
lab0 = -1
l_clock = -1
isFirstChrono = True
isReplay = False
isUnvisibleClock = True
doStopChrono = False
clockX = 0
clockY = 0
posX = 0.0;
posY = 0.0;
posZ = 0.0;
speed = 0.0;

#onButtonClick = [onClickButton0, onClickButton1,
#                 onClickButton2, onClickButton3, onClickButton4]


class UdpHandler(socketserver.BaseRequestHandler):
    def handle(self):
        global appWindow, isReplay

        #self.request is the UDP socket connected to the client
        recData = json.loads(self.request[0].decode("utf-8"))
        if recData["method"] == NEW_APP :
            newApp(recData)
        elif recData["method"] == SET_SIZE :
            setSize(recData)
        elif recData["method"] == SET_TITLE :
            setTitle(recData)
        elif recData["method"] == SET_BACKGROUND_TEXTURE :
            setBackgroundTexture(recData)
        elif recData["method"] == SET_VISIBLE :
            setVisible(recData)
        elif recData["method"] == ADD_LABEL :
            addLabel(recData)
        elif recData["method"] == SET_POSITION :
            setPosition(recData)
        elif recData["method"] == DRAW_BACKGROUND :
            drawBackground(recData)
        elif recData["method"] == SET_BACKGROUND_COLOR :
            setBackgroundColor(recData)
        elif recData["method"] == SET_BACKGROUND_OPACITY :
            setBackgroundOpacity(recData)
        elif recData["method"] == ADD_BUTTON :
            addButton(recData)
        elif recData["method"] == SET_FONT_ALIGNMENT :
            setFontAlignment(recData)
        elif recData["method"] == SET_FONT_COLOR :
            setFontColor(recData)
        elif recData["method"] == SET_FONT_SIZE :
            setFontSize(recData)
        elif recData["method"] == SET_TEXT :
            setText(recData)
        elif recData["method"] == CUSTOM1 :
            startChrono(recData)
        elif recData["method"] == GET_TRACK_CONFIGURATION :
            getTrackConfiguration()
        elif recData["method"] == CUSTOM2 :
            ac.console("stop chrono ...")
            stopChrono(recData)
        elif recData["method"] == CUSTOM3 :
            ac.console("mode replay ...")
            isReplay = True
        elif recData["method"] == CUSTOM4 :
            ac.console("End mode replay ...")
            isReplay = False

def acMain(ac_version):   
    global server, appWindow, lightTexture
    appWindow = ac.newApp("Codriver")
    
    # Start the c# Client
    runEngine = "apps/python/Codriver/bin/Codriver.exe"
    
    p = subprocess.Popen([runEngine])

    acObj["appWindow"] = appWindow
    ac.setSize(appWindow, 50, 20)
    ac.log("AcConnect started ...")
    HOST, PORT = "localhost", 11367
    server = socketserver.UDPServer((HOST, PORT), UdpHandler)
    serverWorkerThread = threading.Thread(target=server.serve_forever)
    serverWorkerThread.daemon = True
    serverWorkerThread.start()
    ac.addOnAppActivatedListener(appWindow, onAppActivated)
    ac.addOnAppDismissedListener(appWindow, onAppDismissed)

    ac.addRenderCallback(appWindow , onFormRender)
    
    ac.log("AcConnect end acMain ...")

    return "Codriver"

def acUpdate(deltaT):
    global server, onOne, posX, posY, posZ, speed, isReplay
    if isReplay :
        res = ac.getCarState(0, acsys.CS.WorldPosition)
        posX = float(res[0])
        posY = float(res[1])
        posZ = float(res[2])
        speed = ac.getCarState(0, acsys.CS.SpeedKMH)
        onPosition(appWindow)

def acShutdown():
    global server
    event = {}
    event["eventtype"] = ON_SHUTDOWN
    sendEvent(event)
    time.sleep(0.5)
    server.shutdown()

###############################################################################
### AC Method
###############################################################################
def newApp(rec) :
    event = {}
    event["eventtype"] = ON_APP_INITIALIZED
    sendEvent(event)

def setSize(rec) :
    sName = rec["control"]
    control = acObj[sName]
    width = rec["ints"][0]
    height = rec["ints"][1]
    ac.setSize(control, width, height)    

def setPosition(rec) :
    sName = rec["control"]
    control = acObj[sName]
    x = rec["ints"][0]
    y = rec["ints"][1]
    ac.setPosition(control, x, y)

def setBackgroundOpacity(rec) :
    sName = rec["control"]
    control = acObj[sName]
    value = rec["floats"][0]
    ac.setBackgroundOpacity(control, value)

def setBackgroundColor(rec) :
    sName = rec["control"]
    control = acObj[sName]
    red = rec["floats"][0]
    blue = rec["floats"][1]
    green = rec["floats"][2]
    ac.setBackgroundColor(control, red, blue, green)

def setVisible(rec) :
    sName = rec["control"]
    control = acObj[sName]
    isVisible = rec["ints"][0]
    ac.setVisible(control, isVisible)

def drawBackground(rec) :
    sName = rec["control"]
    control = acObj[sName]
    value = rec["ints"][0]
    ac.drawBackground(control, value)

def setTitle(rec) :
    sName = rec["control"]
    control = acObj[sName]
    title = rec["texts"][0]
    ac.setTitle(control, title)

def setText(rec) :
    sName = rec["control"]
    control = acObj[sName]
    txt = rec["texts"][0]
    ac.setText(control, txt)
    

def setFontAlignment(rec) :
    sName = rec["control"]
    control = acObj[sName]
    align = rec["texts"][0]
    ac.setFontAlignment(control, align)

def setFontSize(rec) :
    sName = rec["control"]
    control = acObj[sName]
    size = rec["ints"][0]
    ac.setFontSize(control, size)

def setFontColor(rec) :
    sName = rec["control"]
    control = acObj[sName]
    alpha = rec["ints"][0]
    red = rec["ints"][1]
    green = rec["ints"][2]
    blue = rec["ints"][3]
    ac.setFontColor(control, red, green, blue, alpha)

def setBackgroundTexture(rec) :
    sName = rec["control"]
    control = acObj[sName]
    path = rec["texts"][0]
    ac.setBackgroundTexture(control, path)

def getTrackConfiguration() :
    trackConfiguration = ac.getTrackConfiguration(0)
    trackName = ac.getTrackName(0)
    event = {}
    event["eventtype"] = ON_GET_TRACK_CONFIGURATION
    event["trackconfiguration"] = trackConfiguration + "/" + trackName
    sendEvent(event)


def addLabel(rec) :
    sName = rec["control"]
    control = acObj[sName]
    labelName = rec["texts"][0]
    text = rec["texts"][1]
    acObj[labelName] = ac.addLabel(control, text)

def addButton(rec) :
    global iButtonClick, buttonList, lightTexture, lab0
    sName = rec["control"]
    control = acObj[sName]
    buttonName = rec["texts"][0]
    text = rec["texts"][1]
    acObj[buttonName] = ac.addButton(control, text)
    buttonList.append(buttonName)


    if (iButtonClick == 0) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton0)
    elif (iButtonClick == 1) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton1)
    elif (iButtonClick == 2) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton2)
    elif (iButtonClick == 3) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton3)
    elif (iButtonClick == 4) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton4)
    elif (iButtonClick == 5) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton5)
    elif (iButtonClick == 6) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton6)
    elif (iButtonClick == 7) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton7)
    elif (iButtonClick == 8) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton8)
    elif (iButtonClick == 9) :
        iButtonClick += 1
        ac.addOnClickedListener(acObj[buttonName], onClickButton9)

def onAppActivated(appId):
    event = {}
    event["eventtype"] = ON_APP_ACTIVATED
    sendEvent(event)

def onAppDismissed(appId):
    event = {}
    event["eventtype"] = ON_APP_DISMISSED
    sendEvent(event)

def onPosition(appId):
    global posX, posY, posZ
    event = {}
    event["eventtype"] = ON_POSITION
    event["posX"] = posX;
    event["posY"] = posY;
    event["posZ"] = posZ;
    event["speed"] = speed;
    sendEvent(event)



def sendEvent(event) :
    data = json.dumps(event)
    sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    sock.sendto(bytes(data + "\n", "utf-8"), ("localhost", 11366))

###############################################################################
### On Button click 
###############################################################################
def onClickButton0(value1, value2) :
    global buttonList
    SendEventButton(buttonList[0])
def onClickButton1(value1, value2) :
    SendEventButton(buttonList[1])
def onClickButton2(value1, value2) :
    SendEventButton(buttonList[2])
def onClickButton3(value1, value2) :
    SendEventButton(buttonList[3])
def onClickButton4(value1, value2) :
    SendEventButton(buttonList[4])
def onClickButton5(value1, value2) :
    SendEventButton(buttonList[5])
def onClickButton6(value1, value2) :
    SendEventButton(buttonList[6])
def onClickButton7(value1, value2) :
    SendEventButton(buttonList[7])
def onClickButton8(value1, value2) :
    SendEventButton(buttonList[8])
def onClickButton9(value1, value2) :
    SendEventButton(buttonList[9])

def SendEventButton(buttonName) :
    winsound.PlaySound("apps/python/Codriver/Sounds/gui_click.wav", winsound.SND_FILENAME + winsound.SND_ASYNC)
    ac.log("click button " + buttonName )
    event = {}
    event["eventtype"] = ON_BUTTON_CLICK
    event["button"] = buttonName
    sendEvent(event)

###############################################################################
### Specific Codriver
###############################################################################
def stopChrono(rec) :
    global doStopChrono
    doStopChrono = True

def startChrono(rec) :
    global appWindow, lightTexture, lab0, l_clock, isFirstChrono, isUnvisibleClock, doStopChrono,  clockX, clockY
    doStopChrono = False
    clockPosition = int(rec["ints"][6])
    if clockPosition == 0 :
        clockX = 400
        clockY = 30
    if clockPosition == 1 :
        clockX = -260
        clockY = 30
    if clockPosition == 2 :
        clockX = 70
        clockY = -170
    if clockPosition == 3 :
        clockX = 70
        clockY = 220

    if isFirstChrono :
        l_clock = ac.addLabel(appWindow, "");
        ac.setSize(l_clock, 251, 154)
        ac.setPosition(l_clock, clockX, clockY)
        #ac.setPosition(l_clock, 450, 20)
        ac.setBackgroundTexture(l_clock, "apps/python/Codriver/images/FIA Clock.bmp");
        lab0 = ac.addLabel(appWindow, "")
        lightTexture = ac.newTexture(lab0, "apps/python/Codriver/images/red.bmp")

    else :
        ac.setVisible(l_clock, 1)
        lightTexture = ac.newTexture(lab0, "apps/python/Codriver/images/red.bmp")

    isUnvisibleClock = False

    timeStart = datetime.datetime(int(rec["ints"][0]),
                         int(rec["ints"][1]),
                         int(rec["ints"][2]),
                         int(rec["ints"][3]),
                         int(rec["ints"][4]),
                         int(rec["ints"][5]),
                         0)
    countDownThread = threading.Thread(target=countdown, args=(timeStart, ))
    countDownThread.start()


def countdown(timeStart) :
    global appWindow, lightTexture, lab0, l_clock, isUnvisibleClock, doStopChrono, clockX, clockY
    ac.console(str(timeStart))
    is30Sounded = False
    is15Sounded = False
    is10Sounded = False
    is5Sounded = False

    ac.setVisible(lab0, 1)
    ac.setPosition(lab0, clockX + 155, clockY + 70)
    ac.setFontSize(lab0, 70)
    ac.setFontColor(lab0,0.8 , 0.1, 0.1, 1)
    ac.setText(lab0, "")
      
    yellowTexture = ac.newTexture("apps/python/Codriver/images/yellow.bmp")
    greenTexture = ac.newTexture("apps/python/Codriver/images/green.bmp")
    redTexture = ac.newTexture("apps/python/Codriver/images/red.bmp")
    blackTexture = ac.newTexture("apps/python/Codriver/images/black.bmp")

    timeDelta = timeStart - datetime.datetime.utcnow()
    isFalseStart = False

    while True :
        speed = ac.getCarState(0, acsys.CS.SpeedKMH)
        if speed > 0.5 and isFalseStart == False :
            isFalseStart = True
            winsound.PlaySound("apps/python/Codriver/Sounds/False.wav", winsound.SND_FILENAME + winsound.SND_ASYNC)
            event = {}
            event["eventtype"] = ON_FALSE_START
            sendEvent(event)

        sec = timeDelta.seconds + timeDelta.microseconds / 1000000.0 + 1
        if doStopChrono :
            ac.setVisible(l_clock, 0)
            ac.setVisible(lab0, 0)
            isUnvisibleClock = True
            break;
        if 16000 > sec and sec > 0 :
            sec = timeDelta.seconds + timeDelta.microseconds / 1000000.0
            if sec > 5 : 
                lightTexture = redTexture
                ac.setText(lab0, str.format("{:0>2d}".format(ceil(sec))))
                if ceil(sec) == 10 and is10Sounded == False :

                    event = {}
                    event["eventtype"] = ON_10_SEC
                    sendEvent(event)
                    is10Sounded = True
                    winsound.PlaySound("apps/python/Codriver/Sounds/beep.wav", winsound.SND_FILENAME + winsound.SND_ASYNC)
                
                elif ceil(sec) == 30 and is30Sounded == False :
                    event = {}
                    event["eventtype"] = ON_30_SEC
                    sendEvent(event)                    
                    is30Sounded = True

                elif ceil(sec) == 15 and is15Sounded == False :
                    event = {}
                    event["eventtype"] = ON_15_SEC
                    sendEvent(event)                    
                    is15Sounded = True

            else :
                if ceil(sec) == 5 and is5Sounded == False :
                    event = {}
                    event["eventtype"] = ON_5_SEC
                    sendEvent(event)                    
                    is5Sounded = True

                if timeDelta.microseconds > 500000 :

                    if lightTexture != yellowTexture : 
                        lightTexture = yellowTexture
                        ac.setText(lab0, str.format("{:0>2d}".format(ceil(sec))))
                        winsound.PlaySound("apps/python/Codriver/Sounds/beep.wav", winsound.SND_FILENAME + winsound.SND_ASYNC)
                else : 
                    lightTexture = blackTexture
                    ac.setText(lab0, str.format("{:0>2d}".format(ceil(sec))))
        else :
            sec = 0
            winsound.PlaySound("apps/python/Codriver/Sounds/start.wav", winsound.SND_FILENAME + winsound.SND_ASYNC)
            lightTexture = greenTexture
            ac.setText(lab0, str.format("{:0>2d}".format(ceil(sec))))
            time.sleep(2)
            ac.setVisible(l_clock, 0)
            ac.setVisible(lab0, 0)
            isUnvisibleClock = True
            break

        time.sleep(0.050)
        timeDelta = timeStart - datetime.datetime.utcnow()

def onFormRender(deltaT):
    global lightTexture, isUnvisibleClock
    if isUnvisibleClock == False :
        ac.glQuadTextured(clockX + 180, clockY + 45, 30, 30, lightTexture)
